"""
A module implementing interfaces for struct-file generation.
"""

# built-in
from json import dumps
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Dict, NamedTuple

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen import PKG_NAME, VERSION
from ifgen.config import Config

StructConfig = Dict[str, Any]


class GenerateStructTask(NamedTuple):
    """Parameters necessary for struct generation."""

    root: Path
    path: Path
    struct: StructConfig
    config: Dict[str, Any]

    def command(self, command: str, data: str = "", space: str = " ") -> str:
        """Get a command string."""
        return (
            str(self.config["command"])
            + command
            + (space if data else "")
            + data
        )


def create_struct(task: GenerateStructTask) -> None:
    """Create a header file based on a struct definition."""

    with IndentedFileWriter.from_path(task.path, per_indent=4) as writer:
        # Write file header.
        with writer.javadoc():
            writer.write(task.command("file"))
            writer.write(
                task.command("brief", f"Generated by {PKG_NAME} ({VERSION}).")
            )
            writer.empty()
            writer.write(dumps(task.struct, indent=4))

        writer.write("#pragma once")
        writer.empty()

        # Write namespace.
        namespace = "::".join(task.config["namespace"])
        writer.write(f"namespace {namespace}")
        with writer.scope(suffix=f"; // namespace {namespace}"):
            # Write struct definition.
            writer.empty()
            writer.cpp_comment("Testing.")
            writer.empty()


def generate_structs(root: Path, output: Path, config: Config) -> None:
    """Generate struct files."""

    output.mkdir(parents=True, exist_ok=True)

    pool = Pool()  # pylint: disable=consider-using-with
    try:
        pool.map(
            create_struct,
            (
                GenerateStructTask(
                    root, output.joinpath(f"{name}.h"), data, config.data
                )
                for name, data in config.data["structs"].items()
            ),
        )
    finally:
        pool.close()
        pool.join()
